/**
 * 
 */
package de.bstly.board.controller;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLDecoder;
import java.time.Instant;
import java.util.List;
import java.util.Optional;
import java.util.Scanner;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.http.HttpStatus;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

import com.google.common.collect.Lists;

import de.bstly.board.businesslogic.EntryManager;
import de.bstly.board.businesslogic.UserManager;
import de.bstly.board.businesslogic.VoteManager;
import de.bstly.board.controller.support.EntityResponseStatusException;
import de.bstly.board.controller.support.RequestBodyErrors;
import de.bstly.board.controller.validation.EntryValidator;
import de.bstly.board.model.Entry;
import de.bstly.board.model.EntryStatus;
import de.bstly.board.model.RankedEntry;
import de.bstly.board.model.Types;
import de.bstly.board.model.Vote;
import de.bstly.board.model.VoteType;

/**
 * The Class EntryController.
 */
@RestController
@RequestMapping("/entries")
public class EntryController extends BaseController {

	@Autowired
	private EntryManager entryManager;
	@Autowired
	private UserManager userManager;
	@Autowired
	private EntryValidator entryValidator;
	@Autowired
	private VoteManager voteManager;

	@Value("${bstly.board.size:30}")
	private int SIZE;

	/**
	 * Fetch by ranking.
	 *
	 * @param pageParameter the page parameter
	 * @param sizeParameter the size parameter
	 * @param dateParameter the date parameter
	 * @param gravityParameter the gravity parameter
	 * @param ignoreParameter the ignore parameter
	 * @return the page
	 */
	@PreAuthorize("isAuthenticated()")
	@GetMapping()
	public Page<Entry> fetchByRanking(@RequestParam("page") Optional<Integer> pageParameter,
			@RequestParam("size") Optional<Integer> sizeParameter,
			@RequestParam("date") Optional<Instant> dateParameter,
			@RequestParam("gravity") Optional<Double> gravityParameter,
			@RequestParam("ignore") Optional<List<String>> ignoreParameter) {

		if (sizeParameter.isPresent() && sizeParameter.get() > 100) {
			sizeParameter = Optional.of(100);
		}

		Page<RankedEntry> entries = entryManager.fetchByRanking(dateParameter.orElse(Instant.now()),
				gravityParameter.orElse(getGravity()), pageParameter.orElse(0),
				sizeParameter.orElse(SIZE));

		Page<Entry> transformed = new PageImpl<Entry>(
				entries.getContent().stream().map(rankedEntry -> {
					Entry entry = Entry.fromRankedEntry(rankedEntry);
					entry.getMetadata().put("points", rankedEntry.getPoints());
					return entry;
				}).collect(Collectors.toList()), entries.getPageable(), entries.getTotalElements());

		List<String> ignore = ignoreParameter.orElse(Lists.newArrayList());
		entryManager.applyMetadata(getCurrentUsername(), userManager.getKarma(getCurrentUsername()),
				transformed.getContent(), ignore);
		return transformed;
	}

	/**
	 * Fetch by date.
	 *
	 * @param pageParameter the page parameter
	 * @param sizeParameter the size parameter
	 * @param dateParameter the date parameter
	 * @param ignoreParameter the ignore parameter
	 * @return the page
	 */
	@PreAuthorize("isAuthenticated()")
	@GetMapping("/new")
	public Page<Entry> fetchByDate(@RequestParam("page") Optional<Integer> pageParameter,
			@RequestParam("size") Optional<Integer> sizeParameter,
			@RequestParam("date") Optional<Instant> dateParameter,
			@RequestParam("ignore") Optional<List<String>> ignoreParameter) {

		if (sizeParameter.isPresent() && sizeParameter.get() > 100) {
			sizeParameter = Optional.of(100);
		}

		Page<Entry> entries = entryManager.fetchByDate(dateParameter.orElse(Instant.now()),
				pageParameter.orElse(0), sizeParameter.orElse(SIZE));
		List<String> ignore = ignoreParameter.orElse(Lists.newArrayList());
		entryManager.applyMetadata(getCurrentUsername(), userManager.getKarma(getCurrentUsername()),
				entries.getContent(), ignore);
		return entries;
	}

	/**
	 * Fetch by comments.
	 *
	 * @param pageParameter the page parameter
	 * @param sizeParameter the size parameter
	 * @param dateParameter the date parameter
	 * @param gravityParameter the gravity parameter
	 * @param ignoreParameter the ignore parameter
	 * @return the page
	 */
	@PreAuthorize("isAuthenticated()")
	@GetMapping("/comments")
	public Page<Entry> fetchByComments(@RequestParam("page") Optional<Integer> pageParameter,
			@RequestParam("size") Optional<Integer> sizeParameter,
			@RequestParam("date") Optional<Instant> dateParameter,
			@RequestParam("gravity") Optional<Double> gravityParameter,
			@RequestParam("ignore") Optional<List<String>> ignoreParameter) {

		if (sizeParameter.isPresent() && sizeParameter.get() > 100) {
			sizeParameter = Optional.of(100);
		}

		Page<RankedEntry> entries = entryManager.fetchByComments(
				dateParameter.orElse(Instant.now()), gravityParameter.orElse(getGravity()),
				pageParameter.orElse(0), sizeParameter.orElse(SIZE));

		Page<Entry> transformed = new PageImpl<Entry>(
				entries.getContent().stream().map(rankedEntry -> Entry.fromRankedEntry(rankedEntry))
						.collect(Collectors.toList()),
				entries.getPageable(), entries.getTotalElements());

		List<String> ignore = ignoreParameter.orElse(Lists.newArrayList());
		entryManager.applyMetadata(getCurrentUsername(), userManager.getKarma(getCurrentUsername()),
				transformed.getContent(), ignore);
		return transformed;
	}

	/**
	 * Fetch by last.
	 *
	 * @param pageParameter the page parameter
	 * @param sizeParameter the size parameter
	 * @param dateParameter the date parameter
	 * @param ignoreParameter the ignore parameter
	 * @return the page
	 */
	@PreAuthorize("isAuthenticated()")
	@GetMapping("/last")
	public Page<Entry> fetchByLast(@RequestParam("page") Optional<Integer> pageParameter,
			@RequestParam("size") Optional<Integer> sizeParameter,
			@RequestParam("date") Optional<Instant> dateParameter,
			@RequestParam("ignore") Optional<List<String>> ignoreParameter) {

		if (sizeParameter.isPresent() && sizeParameter.get() > 100) {
			sizeParameter = Optional.of(100);
		}

		Page<Entry> entries = entryManager.fetchByLastComment(dateParameter.orElse(Instant.now()),
				pageParameter.orElse(0), sizeParameter.orElse(SIZE));

		List<String> ignore = ignoreParameter.orElse(Lists.newArrayList());
		entryManager.applyMetadata(getCurrentUsername(), userManager.getKarma(getCurrentUsername()),
				entries.getContent(), ignore);
		return entries;
	}

	/**
	 * Fetch by user.
	 *
	 * @param username the username
	 * @param pageParameter the page parameter
	 * @param sizeParameter the size parameter
	 * @param dateParameter the date parameter
	 * @param ascParameter the asc parameter
	 * @param ignoreParameter the ignore parameter
	 * @return the page
	 */
	@PreAuthorize("isAuthenticated()")
	@GetMapping("/byuser/{username}")
	public Page<Entry> fetchByUser(@PathVariable("username") String username,
			@RequestParam("page") Optional<Integer> pageParameter,
			@RequestParam("size") Optional<Integer> sizeParameter,
			@RequestParam("date") Optional<Instant> dateParameter,
			@RequestParam("asc") Optional<Boolean> ascParameter,
			@RequestParam("ignore") Optional<List<String>> ignoreParameter) {

		if (sizeParameter.isPresent() && sizeParameter.get() > 100) {
			sizeParameter = Optional.of(100);
		}

		Page<Entry> entries = entryManager.fetchByUser(username,
				dateParameter.orElse(Instant.now()), pageParameter.orElse(0),
				sizeParameter.orElse(SIZE), ascParameter.orElse(false));
		List<String> ignore = ignoreParameter.orElse(Lists.newArrayList());
		entryManager.applyMetadata(getCurrentUsername(), userManager.getKarma(getCurrentUsername()),
				entries.getContent(), ignore);
		return entries;
	}

	/**
	 * Gets the entry.
	 *
	 * @param id the id
	 * @param ignoreParameter the ignore parameter
	 * @return the entry
	 */
	@PreAuthorize("isAuthenticated()")
	@GetMapping("/entry/{id}")
	public Entry getEntry(@PathVariable("id") Long id,
			@RequestParam("ignore") Optional<List<String>> ignoreParameter) {
		Entry entry = entryManager.get(id);

		if (entry == null) {
			throw new EntityResponseStatusException(HttpStatus.NOT_FOUND);
		}

		List<String> ignore = ignoreParameter.orElse(Lists.newArrayList());
		entryManager.applyMetadata(getCurrentUsername(), userManager.getKarma(getCurrentUsername()),
				entry, ignore);

		return entry;
	}

	/**
	 * Creates the entry.
	 *
	 * @param entry the entry
	 * @param ignoreParameter the ignore parameter
	 * @return the entry
	 */
	@PreAuthorize("isAuthenticated()")
	@PostMapping()
	public Entry createEntry(@RequestBody Entry entry,
			@RequestParam("ignore") Optional<List<String>> ignoreParameter) {
		RequestBodyErrors bindingResult = new RequestBodyErrors(entry);
		entryValidator.validate(entry, bindingResult);

		if (bindingResult.hasErrors()) {
			throw new EntityResponseStatusException(bindingResult.getAllErrors(),
					HttpStatus.UNPROCESSABLE_ENTITY);
		}

		entry.setCreated(Instant.now());
		entry.setAuthor(getCurrentUsername());
		entry.setEntryStatus(EntryStatus.NORMAL);
		entry = entryManager.save(entry);

		Vote vote = new Vote();
		vote.setTarget(entry.getId());
		vote.setType(VoteType.up);
		vote.setTargetType(Types.entry);
		vote.setAuthor(getCurrentUsername());
		voteManager.save(vote);

		List<String> ignore = ignoreParameter.orElse(Lists.newArrayList());
		entryManager.applyMetadata(getCurrentUsername(), userManager.getKarma(getCurrentUsername()),
				entry, ignore);

		return entry;
	}

	/**
	 * Gets the title.
	 *
	 * @param url the url
	 * @return the title
	 */
	@PreAuthorize("isAuthenticated()")
	@GetMapping("/helper/title")
	public String getTitle(@RequestParam("url") String url) {
		InputStream response = null;
		Scanner scanner = null;
		try {
			response = new URL(URLDecoder.decode(url, "utf-8")).openStream();
			scanner = new Scanner(response);
			String responseBody = scanner.useDelimiter("\\A").next();
			return responseBody.substring(responseBody.indexOf("<title>") + 7,
					responseBody.indexOf("</title>"));
		} catch (IOException ex) {
		} finally {
			try {
				if (response != null) {
					response.close();
				}
				if (scanner != null) {
					scanner.close();
				}
			} catch (IOException ex) {
			}
		}
		return "";
	}

}
